"""Class for Substance.

Substance class serves as the (abstract) base class for `Element`, `Molecule` and
`Mixture`.
"""

from __future__ import annotations

from collections import defaultdict
from typing import Any, TypeAlias, Union, Literal, Iterable
from abc import ABCMeta, abstractmethod
import copy

from .constants import N_A
from .conversion import at_to_wt, wt_to_at, vol_to_at, at_to_vol
from .isotope import Isotope
from .node import Node, char_sets


Constituent: TypeAlias = Union["Substance", Isotope]

class Substance(metaclass=ABCMeta):
    """Abstract base class for all types of substances.

    Each substance has a name and is a composition of constituents, each
    accounting for a certain fraction. Internally, this is saved as atomic (or
    mole) fraction. Every substance has a molar mass, that can be calculated
    from its composition, since the basis is made of isotopes, that have their 
    mass implemented. Density can generally not be calculated but instead must
    be stated explicitly. Optionally, a shorter name can be given in the form
    of a symbol.
    """
    
    @classmethod
    @abstractmethod
    def _get_allowed_constituents(cls):
        """Returns a tuple of allowed classes for constituents."""
        return NotImplementedError
    
    _NORMALISE = True  # normalisation of atomic fractions
    _RIGHT_ALIGN_PREF = None  # right alignment preference of substance in tree prints

    @abstractmethod
    def __init__(self, name: str, composition: dict[Constituent, float], mode: str = "_legacy", **kwargs) -> None:
        """Constructor of substance.
        
        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps a constituent of the substance to its
                fraction. The fractions are physically interpreted according
                to the given mode. Values are normalsied and don't need to add
                up to unity.
            mode:
                Fractions in the composition can be interpreted as atomic,
                weight or volumetric fractions. Volumetric is only valid, if
                all constituents have a density. The default value allows
                inputs as before v1.1.0, where positive fractions refer to
                atomic and negative fractions to weight.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.

        Raises:
            ValueError:
                If non-valid constructor mode or given constituent is not
                allowed as constituent for substance.
        """

        self._name = name                             # name of the substance
        self._composition: dict[Constituent, float]   # {constituent: atomic (mole) fraction}
        self._M: float                                # molar mass [g mol^-1]
        self._rho: float                              # density [g cm^-3]
        self._symbol: str                             # symbol of the substance
        self._data: dict[str, float]                  # additional data for storage or calculation of mean values

        # ensure input constituents are of allowed classes
        for constituent in composition.keys():
            if not self.instance_is_allowed(constituent):
                type_names = ", ".join(_type.__name__ for _type in self._get_allowed_constituents())
                raise ValueError(f"Could not create {self.__class__.__name__} object with type {constituent.__class__.__name__} as constituent.\n"
                                 + f"Valid classes for {self.__class__.__name__} constituents: {type_names}.")

        # ensure atomic fraction of input composition
        if mode == "_legacy":  # as done before 1.1.0
            composition = self._inp_composition_legacy(composition)
        elif mode in {"atomic", "at", "mole", "mol"}:  # no conversion, only assure positive values
            composition = self._inp_composition_atomic(composition)
        elif mode in {"weight", "wt"}:  # convert from weight fractions
            composition = self._inp_composition_weight(composition)
        elif mode in {"volume", "vol"}:  # convert from volume fractions
            composition = self._inp_composition_volume(composition)
        elif mode == "_skip":  # no action
            composition = composition
        else:
            raise ValueError(f"Unknown constructor mode \"{mode}\".")

        # populate composition dictionary
        x_sum = sum(composition.values())
        if self._NORMALISE:  # normalise fractions
            self._composition = {constituent: (x_i / x_sum) for constituent, x_i in composition.items() if x_i > 0}
        else:  # no normalisation
            self._composition = {constituent: x_i for constituent, x_i in composition.items() if x_i > 0}

        # extract variables from kwargs (or calculate)
        self._M = kwargs.get("M", self._calc_M())
        self._rho = kwargs.get("rho", 0.0)
        self._symbol = kwargs.get("symbol", "")
        self._data = kwargs.get("data", dict())
    
    @classmethod
    def _inp_composition_atomic(cls, inp_composition: dict[Constituent, float]) -> dict[Constituent, float]:
        """Returns constituent dictionary with atomic (mole) fractions for given atomic (mole) fractions."""
        return {constituent: abs(x_i) for constituent, x_i in inp_composition.items()}
    
    @classmethod
    def _inp_composition_weight(cls, inp_composition: dict[Constituent, float]) -> dict[Constituent, float]:
        """Returns constituent dictionary with atomic (mole) fractions for given weight fractions."""
        
        wt_fracs = [abs(w_i) for w_i in inp_composition.values()]
        molar_masses = [constituent.M for constituent in inp_composition.keys()]

        at_fracs = wt_to_at(wt_fracs, molar_masses)
        return {constituent: x_i for constituent, x_i in zip(inp_composition.keys(), at_fracs)}
    
    @classmethod
    def _inp_composition_volume(cls, inp_composition: dict[Constituent, float]) -> dict[Constituent, float]:
        """Returns constituent dictionary with atomic (mole) fractions for given volume fractions."""
        
        vol_fracs = [abs(phi_i) for phi_i in inp_composition.values()]
        molar_volumes = [constituent.V_m for constituent in inp_composition.keys()]

        at_fracs = vol_to_at(vol_fracs, molar_volumes)
        return {constituent: x_i for constituent, x_i in zip(inp_composition.keys(), at_fracs)}

    @classmethod
    def _inp_composition_legacy(cls, inp_composition: dict[Constituent, float]) -> dict[Constituent, float]:
        """Returns constituent dictionary with atomic (mole) fractions as done before 1.1.0.
        
        Legacy method to maintain backwards compatibility.
        """
        
        if not all(frac > 0 for frac in inp_composition.values()):  # either all negative or mixed
            if all(frac < 0 for frac in inp_composition.values()):  # all negative -> weight fractions given, need to be converted
                return cls._inp_composition_weight(inp_composition)
            else:  # mixed, not allowed
                raise ValueError(f"Could not create {cls.__name__}: Mixing of atomic and weight fractions is not allowed.")
        else:  # all positive -> atomic fractions given, no action needed
            return inp_composition


    # ########
    # Factory Constructor
    # ########

    @classmethod
    def from_atomic(cls, name: str, composition: dict[Constituent, float], **kwargs) -> Substance:
        """Constructor for providing atomic (mole) fractions of constituents.

        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps a constituent of the substance to its
                atomic fraction. Values are normalsied and don't need to add up
                to unity.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.

        Returns:
            Substance instance.
        """
        return cls(name, composition, mode="atomic", **kwargs)

    @classmethod
    def from_weight(cls, name: str, composition: dict[Constituent, float], **kwargs) -> Substance:
        """Constructor for providing weight fractions of constituents.

        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps a constituent of the substance to its
                weight fraction. Values are normalsied and don't need to add up
                to unity.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.
                
        Returns:
            Substance instance.
        """
        return cls(name, composition, mode="weight", **kwargs)

    @classmethod
    def from_volume(cls, name: str, composition: dict[Constituent, float], **kwargs) -> Substance:
        """Constructor for providing volume fractions of constituents.
        
        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps a constituent of the substance to its
                volume fraction. Values are normalsied and don't need to add up
                to unity.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.
                
        Returns:
            Substance instance.
        """
        return cls(name, composition, mode="volume", **kwargs)
    

    def copy(
            self,
            name: str = None,
            M: float = None,
            rho: float = None,
            symbol: str = None,
            data: dict[str, float] = None
        ) -> Substance:
        """Copy substance and edit properties.
        
        It is not possible to edit the composition. To do so, create a new
        substance.

        Args:
            name:
                Descriptive name.
            M:
                Molar mass.
            rho:
                Density.
            symbol:
                Short symbol.
            data:
                Dictionary with additional data to store or use for mean values
                in mixtures.
        
        Returns:
            Copied substance instance.
        """

        new = copy.deepcopy(self)

        if name:
            new._name = name
        if M:
            new._M = M
        if rho:
            new._rho = rho
        if symbol:
            new._symbol = symbol
        if data:
            new._data = data

        return new


    # ########
    # Properties
    # ########

    @property
    def name(self):
        """Given name."""
        return self._name

    @property
    def composition(self):
        """Dictionary of constituents and their atomic (mole) fraction."""
        return self._composition

    @property
    def M(self):
        """Molar mass [g mol^-1]."""
        return self._M
    
    @property
    def rho(self):
        """Density [g cm^-3]."""
        return self._rho

    @property
    def symbol(self):
        """Symbol."""
        return self._symbol
    
    @property
    def data(self):
        """Additional data dictionary for storage or calculation of mean values."""
        return self._data

    @property
    def V_m(self):
        """Molar volume [cm^3 mol^-1]."""
        return self._calc_V_m()
    
    @property
    def n(self):
        """Number density [cm^-3]."""
        return self._calc_n()


    # ########
    # Quantity Calculation
    # ########

    def _calc_M(self) -> float:
        r"""Calculates average molar mass.
        
        The molar masses of all constituents are weighted by their atomic
        fraction and summed up:
            $$\overline{M} = \sum_i \left( x_i \cdot M_i \right)$$
        Will return zero if calculation is not possible.
        """

        if all(constituent.M > 0 for constituent in self._composition.keys()):
            return sum(x_i*constituent.M for constituent, x_i in self._composition.items())
        else:
            return 0.0

    def _calc_V_m(self) -> float:
        r"""Calculates molar volume.

        Molar mass is divided by density:
            $$V_m = \frac{M}{\rho}$$
        Will return zero if calculation is not possible.
        """

        if self._M != 0 and self._rho != 0:
            return self._M / self._rho
        else:
            return 0.0
        
    def _calc_n(self) -> float:
        r"""Calculates number density.

        Avogadro constant divided by molar volume (density is normalised by the
        mass of a single atom or molecule):
            $$n = \frac{N_{\mathrm{A}}}{V_m} = \frac{N_{\mathrm{A}}}{M} \rho$$
        Will return zero if calculation is not possible.
        """

        if self._M != 0 and self._rho != 0:
            return N_A / self._M * self._rho
        else:
            return 0.0

    
    # ########
    # Conversion
    # ########

    def get_composition_in_wt(self) -> dict[Constituent, float]:
        """Returns constituents with their weight fractions."""
        
        at_fracs = [x_i for x_i in self._composition.values()]
        molar_masses = [constituent.M for constituent in self._composition.keys()]

        wt_fracs = at_to_wt(at_fracs, molar_masses)
        return {constituent: w_i for constituent, w_i in zip(self._composition.keys(), wt_fracs)}
   
    def get_composition_in_vol(self) -> dict[Constituent, float]:
        """Returns constituents with their volume fractions."""
        
        at_fracs = [x_i for x_i in self._composition.values()]
        molar_volumes = [constituent.V_m for constituent in self._composition.keys()]

        vol_fracs = at_to_vol(at_fracs, molar_volumes)
        return {constituent: phi_i for constituent, phi_i in zip(self._composition.keys(), vol_fracs)}
        

    # ########
    # Collection
    # ########
    
    def _append_elements(
            self, element_list: dict[int, tuple[Substance, float]],
            by_weight: bool = False, f_p: float = 1.0
        ) -> dict[int, tuple[Substance, float]]:
        """Collects all contained elements.
        
        Method is called upon each constituent, until `Element`s are extracted.
        The fraction of the parent is multiplied onto the
        fraction of the constituent. Atomic and weight fractions are possible
        to be fetched. The element, its parent and the fraction of the element
        in the parent are used to generate a (unique) id. Fractions need to be
        normalised.

        Args:
            element_list:
                Container that collects all elements. Gets passed downwards.
                Index "-1" is reserved for counting.
            by_weight:
                Flag to fetch weight fractions of constituents.
            f_p:
                Parent fraction, that is multiplied onto constituent fractions.
            parent:
                Parent object used for generating id.
        
        Returns:
            Dictionary that maps id to element with its fraction
        """

        element_list[-1] = element_list[-1] + 1  # increase counter

        if not by_weight:
            composition = self._composition
        else:
            composition = self.get_composition_in_wt()  
        
        for constituent, f_i in composition.items():
            constituent._append_elements(element_list, by_weight, f_p*f_i)

        return element_list
    
    def _elemental_composition(
            self, by_weight: bool = False
        ) -> dict[int, tuple[Substance, float]]:
        """Collects all contained elements with their normalised fraction.
        
        Args:
            by_weight:
                Flag to fetch weight fractions of constituents.
            
        Returns:
            Dictionary that maps id to element with its fraction
        """

        # get raw elemental composition
        gathered_elements = self._append_elements({-1: -1}, by_weight)
        items = gathered_elements.pop(-1)  # remove counter
        
        # get sum for normalisation
        _, fractions = list(map(list, zip(*gathered_elements.values())))
        norm_tmp = sum(fractions)
        
        return {ident: (element, fraction/norm_tmp) for ident, (element, fraction) in gathered_elements.items()}

    def get_elements(self, mode: Literal["atomic", "weight"] = "atomic"):
        """Returns dict of all contained elements with their summed fraction.
        
        Args:
            mode:
                Wether 'atomic' or 'weight' fractions are to be fetched.
        
        Returns:
            Dictionary that maps occuring elements to their fraction.
        """

        if mode in {"atomic", "at", "mole", "mol"}:
            gathered_elements = self._elemental_composition(by_weight=False)
        elif mode in {"weight", "wt"}:
            gathered_elements = self._elemental_composition(by_weight=True)
        else:
            raise ValueError(f"Mode \"{mode}\" not supported for gathering elements.")
        
        elements = defaultdict(list)
        for element, fraction in gathered_elements.values():
            elements[element].append(fraction)

        return {element: sum(fracs) for element, fracs in sorted(elements.items())}
    
    def _isotopic_composition(
            self, by_weight: bool = False,
            use_natural: bool | Iterable = False
        ) -> dict[int, tuple[Isotope, float]]:
        """Collects all contained isotopes with their normalised fraction.
        
        Args:
            by_weight:
                Flag to fetch weight fractions of constituents.
            
        Returns:
            Dictionary that maps id to isotope with its fraction
        """

        elemental_composition = self._elemental_composition(by_weight)
        isotopic_composition = {}

        def append_isotopes(id: int, element: Substance, fraction: float):
            if not by_weight:
                composition = element._composition
            else:
                composition = element.get_composition_in_wt()
            for j, (isotope, iso_fraction) in enumerate(composition.items(), start=1):
                isotopic_composition[id+j] = (isotope, fraction*iso_fraction)

        def append_surrogate(id: int, element: Substance, fraction: float):
            surrogate = element.surrogate_isotope()
            if surrogate:
                isotopic_composition[id] = (surrogate, fraction)
            else:
                append_isotopes(id, element, fraction)

        for id, (element, fraction) in elemental_composition.items():
            if use_natural:
                if isinstance(use_natural, Iterable):  # have to check if considered
                    if element in use_natural:
                        append_surrogate(id, element, fraction)
                    else:
                        append_isotopes(id, element, fraction)
                else:  # must be boolean true, use elemental composition
                    append_surrogate(id, element, fraction)
            else:
                append_isotopes(id, element, fraction)

        return isotopic_composition
    
    def get_isotopes(
            self, mode: Literal["atomic", "weight"] = "atomic", 
            use_natural: bool | Iterable = False
        ) -> dict[Isotope, float]:
        """Returns dict of all contained isotopes with their summed fraction.
        
        Args:
            mode:
                Wether 'atomic' or 'weight' fractions are to be fetched.
            use_natural:
                Flag to use fraction of element, if it is natural.
                Alternatively, a collection of elements can be supplied, that
                shall be considered.
        
        Returns:
            Dictionary that maps occuring isotopes to their fraction.
        """
        
        if mode in {"atomic", "at", "mole", "mol"}:
            gathered_isotopes = self._isotopic_composition(by_weight=False, use_natural=use_natural)
        elif mode in {"weight", "wt"}:
            gathered_isotopes = self._isotopic_composition(by_weight=True, use_natural=use_natural)
        else:
            raise ValueError(f"Mode \"{mode}\" not supported for gathering isotopes.")

        isotopes = defaultdict(list)
        for isotope, fraction in gathered_isotopes.values():
            isotopes[isotope].append(fraction)

        return {isotope: sum(fracs) for isotope, fracs in sorted(isotopes.items())}

        
    # ########
    # Functions
    # ########

    @classmethod
    def instance_is_allowed(cls, instance: Any) -> bool:
        """Returns, if given instane is of an allowed class (or subclass)."""
        return isinstance(instance, cls._get_allowed_constituents())
    
    @classmethod
    def type_is_allowed(cls, dtype: type) -> bool:
        """Returns, if given type is an allowed class."""
        return all(dtype is allowed for allowed in cls._get_allowed_constituents())


    # ########
    # Tree
    # ########        

    def make_node(
            self, tree_mode: Literal["input", "composition"],
            atomic: bool = True, weight: bool = False, volume: bool = False
        ) -> Node:
        """Creates a hierarchical node structure with instance as the root.
        
        Args:
            atomic:
                Get atomic (mole) fraction for each node.
            weight:
                Get weight fraction for each node.
            volume:
                Get volume fraction for each node.
        
        Returns:
            Node structure with instance as root.
        """

        def add_constituents_input(parent_node: Node):

            nonlocal node_i
            substance: Substance = parent_node.content
            constituents = list(substance.composition.keys())

            # decide for fraction to be used
            if atomic:
                at_fracs = list(substance._composition.values())
            if weight:
                try:
                    wt_fracs = list(substance.get_composition_in_wt().values())
                except ValueError as ex:  # constituent has no molar mass
                    wt_fracs = None
            if volume:
                try:
                    vol_fracs = list(substance.get_composition_in_vol().values())
                except (ValueError, AttributeError) as ex:  # constituent has no density or is an isotope
                    vol_fracs = None


            for i, constituent in enumerate(constituents):
                
                node_i += 1

                # create data dictionary
                constituent_data = {}
                if atomic:
                    constituent_data["x"] = at_fracs[i]
                if weight and wt_fracs:
                    constituent_data["w"] = wt_fracs[i]
                if volume and vol_fracs:
                    constituent_data["phi"] = vol_fracs[i]
                if constituent.M:
                    constituent_data["M"] = constituent.M
                try:
                    if constituent.rho:
                        constituent_data["rho"] = constituent.rho
                except AttributeError as ex:
                    pass

                # create node of constituent
                node = Node(label=str(constituent), content=constituent, parent=parent_node, data=constituent_data, id=node_i)

                # create child node for constituents of constituent (if applicable)
                if isinstance(constituent, Isotope):
                    continue
                else:
                    add_constituents_input(parent_node=node)
            
        def add_constituents_composition(parent_node: Node):

            nonlocal node_i
            substance: Substance = parent_node.content
            constituents = list(substance.composition.keys())

            for i, constituent in enumerate(constituents):

                node_i += 1
                
                # create data dictionary
                constituent_data = {}
                if atomic:
                    try:
                        constituent_data["x"] = composition_atomic[node_i][1]
                    except KeyError:
                        pass
                if weight:
                    try:
                        constituent_data["w"] = composition_weight[node_i][1]
                    except KeyError:
                        pass
                if constituent.M:
                    constituent_data["M"] = constituent.M
                try:
                    if constituent.rho:
                        constituent_data["rho"] = constituent.rho
                except AttributeError as ex:
                    pass

                # create node of constituent
                node = Node(label=str(constituent), content=constituent, parent=parent_node, data=constituent_data, id=node_i)
 
                # create child node for constituents of constituent (if applicable)
                if isinstance(constituent, Isotope):
                    node.align_right(level=-1, label_size=6)
                else:
                    node.align_right(level=constituent._RIGHT_ALIGN_PREF)
                    add_constituents_composition(parent_node=node)

        # create data dictionary
        root_data = {}
        if self._M:
            root_data["M"] = self._M
        if self._rho:
            root_data["rho"] = self._rho

        # create root node and start recursive hierarchy construction
        node_i = 0
        root = Node(label=str(self), content=self, data=root_data, id=node_i)
        if tree_mode == "input":
            add_constituents_input(parent_node=root)
        elif tree_mode == "composition":
            # atomic composition
            if atomic:
                composition_atomic = self._elemental_composition(by_weight=False)
                composition_atomic.update(self._isotopic_composition(by_weight=False))
            else:
                composition_atomic = None
            # weight composition
            if weight:
                composition_weight = self._elemental_composition(by_weight=True)
                composition_weight.update(self._isotopic_composition(by_weight=True))
            else:
                composition_weight = None
            # recursive hierarchy construction
            add_constituents_composition(parent_node=root)

        return root
    
    def print_tree_input(
            self, atomic: bool = True, weight: bool = False, volume: bool = False, *,
            char_set: char_sets = "box_drawings_light", **kwargs
        ) -> None:
        """Prints the hierarchical tree structure.
        
        Args:
            atomic:
                Get atomic (mole) fraction for each node.
            weight:
                Get weight fraction for each node.
            volume:
                Get volume fraction for each node.
            char_set:
                Character set that is used to print lines in the tree.
            **kwargs:
                Arguments passed to the plotting routine of Node.

        Keyword Args:
            frac_fmt (str):
                Format string to be used for fractions.
            prop_fmt (str):
                Format string to be used for physical properties.
        """
        self.make_node(
            "input", atomic, weight, volume,
        ).print_tree(char_set=char_set, **kwargs)

    def print_tree_composition(
            self, atomic: bool = True, weight: bool = False, *,
            char_set: char_sets = "box_drawings_light", **kwargs
        ) -> None:
        """Prints the hierarchical tree structure.
        
        Args:
            atomic:
                Get atomic (mole) fraction for each node.
            weight:
                Get weight fraction for each node.
            char_set:
                Character set that is used to print lines in the tree.
            **kwargs:
                Arguments passed to the plotting routine of Node.

        Keyword Args:
            frac_fmt (str):
                Format string to be used for fractions.
            prop_fmt (str):
                Format string to be used for physical properties.
        """
        self.make_node(
            "composition", atomic, weight,
        ).print_tree(char_set=char_set, **kwargs)


    # ########
    # Print
    # ########


    # ########
    # Operators
    # ########

    def __str__(self):
        return self._name

    def __repr__(self):
        return f"{self.__class__.__name__} \"{self._name}\""

    def __hash__(self):
        return hash((self.__class__, self._name))
    
    def __eq__(self, other):
        if type(self) is type(other):
            return self._name == other._name
        else:
            return NotImplemented
        

    # ########
    # Debug
    # ########
    
    def _compare_converted_isotopes(self) -> dict[str, list[float]]:
        """Debug function to test consistency of fraction modes.
        
        Gets isotopes of substance via "atomic" and "weight" mode, then converts
        each case and compares the atomic and weight fractions from each case
        respectively. For correct behaviour, differences should approach zero.
        """

        # extract isotopes in atomic mode and convert to weight
        tmp = self.get_isotopes(mode="atomic")
        iso_with_at = tmp.keys()
        direct_at = list(tmp.values())
        conv_wt = at_to_wt(at_fracs=direct_at, molar_masses=[isotope.M for isotope in iso_with_at])

        # extract isotopes in weight mode and convert to atomic
        tmp = self.get_isotopes(mode="weight")
        iso_with_wt = tmp.keys()
        direct_wt = list(tmp.values())
        conv_at = wt_to_at(wt_fracs=direct_wt, molar_masses=[isotope.M for isotope in iso_with_wt])

        assert(iso_with_at == iso_with_wt)

        differences = defaultdict(list)

        # print for atomic percent
        print(f"Isotope |   direct atomic  | converted atomic | difference atomic (relative difference)")
        for i, isotope in enumerate(iso_with_at):
            dir = direct_at[i]
            conv = conv_at[i]
            dif = dir - conv
            dif_rel = dif / dir
            print(f"{isotope.name:>7} |  {dir:14.12f}  |  {conv:14.12f}  |  {dif: 12.9e} ({dif_rel: 12.9e})")
            differences["at"].append(dif)
            differences["rel_at"].append(dif_rel)

        print()

        # print for weight percent
        print(f"Isotope |   direct weight  | converted weight | difference weight (relative difference)")
        for i, isotope in enumerate(iso_with_wt):
            dir = direct_wt[i]
            conv = conv_wt[i]
            dif = dir - conv
            dif_rel = dif / dir
            print(f"{isotope.name:>7} |  {dir:14.12f}  |  {conv:14.12f}  |  {dif: 12.9e} ({dif_rel: 12.9e})")
            differences["wt"].append(dif)
            differences["rel_wt"].append(dif_rel)
        

        return differences
    


"""Physical constants and element mappings.

Physical constants for calculations with mole.
Mappings of atomic number to element symbol and element symbol to element name.
"""


N_A = 6.02214076E+23
"""Avogadro constant [mol^-1]."""

M_u = 0.99999999965E-03  # ...(30)
"""Molar mass constant [kg mol^-1] (after 2019 redefinition)."""

ATOM_NUMB_TO_SYMBOL = {
    1:   "H",
    2:   "He",
    3:   "Li",
    4:   "Be",
    5:   "B",
    6:   "C",
    7:   "N",
    8:   "O",
    9:   "F",
    10:  "Ne",
    11:  "Na",
    12:  "Mg",
    13:  "Al",
    14:  "Si",
    15:  "P",
    16:  "S",
    17:  "Cl",
    18:  "Ar",
    19:  "K",
    20:  "Ca",
    21:  "Sc",
    22:  "Ti",
    23:  "V",
    24:  "Cr",
    25:  "Mn",
    26:  "Fe",
    27:  "Co",
    28:  "Ni",
    29:  "Cu",
    30:  "Zn",
    31:  "Ga",
    32:  "Ge",
    33:  "As",
    34:  "Se",
    35:  "Br",
    36:  "Kr",
    37:  "Rb",
    38:  "Sr",
    39:  "Y",
    40:  "Zr",
    41:  "Nb",
    42:  "Mo",
    43:  "Tc",
    44:  "Ru",
    45:  "Rh",
    46:  "Pd",
    47:  "Ag",
    48:  "Cd",
    49:  "In",
    50:  "Sn",
    51:  "Sb",
    52:  "Te",
    53:  "I",
    54:  "Xe",
    55:  "Cs",
    56:  "Ba",
    57:  "La",
    58:  "Ce",
    59:  "Pr",
    60:  "Nd",
    61:  "Pm",
    62:  "Sm",
    63:  "Eu",
    64:  "Gd",
    65:  "Tb",
    66:  "Dy",
    67:  "Ho",
    68:  "Er",
    69:  "Tm",
    70:  "Yb",
    71:  "Lu",
    72:  "Hf",
    73:  "Ta",
    74:  "W",
    75:  "Re",
    76:  "Os",
    77:  "Ir",
    78:  "Pt",
    79:  "Au",
    80:  "Hg",
    81:  "Tl",
    82:  "Pb",
    83:  "Bi",
    84:  "Po",
    85:  "At",
    86:  "Rn",
    87:  "Fr",
    88:  "Ra",
    89:  "Ac",
    90:  "Th",
    91:  "Pa",
    92:  "U",
    93:  "Np",
    94:  "Pu",
    95:  "Am",
    96:  "Cm",
    97:  "Bk",
    98:  "Cf",
    99:  "Es",
    100: "Fm",
    101: "Md",
    102: "No",
    103: "Lr",
    104: "Rf",
    105: "Db",
    106: "Sg",
    107: "Bh",
    108: "Hs",
    109: "Mt",
    110: "Ds",
    111: "Rg",
    112: "Cn",
    113: "Nh",
    114: "Fl",
    115: "Mc",
    116: "Lv",
    117: "Ts",
    118: "Og"
}
"""Maps atomic number to element symbol."""

SYMBOL_TO_NAME = {
    "H":  "hydrogen",
    "He": "helium",
    "Li": "lithium",
    "Be": "berylium",
    "B":  "boron",
    "C":  "carbon",
    "N":  "nitrogen",
    "O":  "oxygen",
    "F":  "fluorine",
    "Ne": "neon",
    "Na": "sodium",
    "Mg": "magnesium",
    "Al": "aluminium",
    "Si": "silicon",
    "P":  "phosphorus",
    "S":  "sulphur",
    "Cl": "chlorine",
    "Ar": "argon",
    "K":  "potassium",
    "Ca": "calcium",
    "Sc": "scandium",
    "Ti": "titanium",
    "V":  "vanadium",
    "Cr": "chromium",
    "Mn": "manganese",
    "Fe": "iron",
    "Co": "cobalt",
    "Ni": "nickel",
    "Cu": "copper",
    "Zn": "zinc",
    "Ga": "gallium",
    "Ge": "germanium",
    "As": "arsenic",
    "Se": "selenium",
    "Br": "bromine",
    "Kr": "krypton",
    "Rb": "rubidium",
    "Sr": "strontium",
    "Y":  "yttrium",
    "Zr": "zirconium",
    "Nb": "niobium",
    "Mo": "molybdenum",
    "Tc": "technetium",
    "Ru": "ruthenium",
    "Rh": "rhodium",
    "Pd": "palladium",
    "Ag": "silver",
    "Cd": "cadmium",
    "In": "indium",
    "Sn": "tin",
    "Sb": "antimony",
    "Te": "tellurium",
    "I":  "iodine",
    "Xe": "xenon",
    "Cs": "caesium",
    "Ba": "barium",
    "La": "lanthanum",
    "Ce": "cerium",
    "Pr": "praseodymium",
    "Nd": "neodymium",
    "Pm": "promethium",
    "Sm": "samarium",
    "Eu": "europium",
    "Gd": "gadolinium",
    "Tb": "terbium",
    "Dy": "dysprosium",
    "Ho": "holmium",
    "Er": "erbium",
    "Tm": "thulium",
    "Yb": "ytterbium",
    "Lu": "lutetium",
    "Hf": "hafnium",
    "Ta": "tantalum",
    "W":  "tungsten",
    "Re": "rhenium",
    "Os": "osmium",
    "Ir": "iridium",
    "Pt": "platinum",
    "Au": "gold",
    "Hg": "mercury",
    "Tl": "thallium",
    "Pb": "lead",
    "Bi": "bismuth",
    "Po": "polonium",
    "At": "astatine",
    "Rn": "radon",
    "Fr": "francium",
    "Ra": "radium",
    "Ac": "actinium",
    "Th": "thorium",
    "Pa": "protactinium",
    "U":  "uranium",
    "Np": "neptunium",
    "Pu": "plutonium",
    "Am": "americium",
    "Cm": "curium",
    "Bk": "berkelium",
    "Cf": "californium",
    "Es": "einsteinium",
    "Fm": "fermium",
    "Md": "mendelevium",
    "No": "nobelium",
    "Lr": "lawrencium",
    "Rf": "rutherfordium",
    "Db": "dubnium",
    "Sg": "seaborgium",
    "Bh": "bohrium",
    "Hs": "hassium",
    "Mt": "meitnerium",
    "Ds": "darmstadtium",
    "Rg": "roentgenium",
    "Cn": "copernicium",
    "Nh": "nihonium",
    "Fl": "flerovium",
    "Mc": "moscovium",
    "Lv": "livermorium",
    "Ts": "tennessine",
    "Og": "oganesson"
}
"""Maps element symbol to element long name."""



from .isotope   import Isotope
from .substance import Substance
from .element   import Element
from .molecule  import Molecule
from .mixture   import Mixture

from .isotopes import *
from .elements import *

from .conversion import *


# DEBUG



"""Class for Node, its helper classes and legacy seperator strings.

The Node class is used to create tree structures of `Substance` objects.
The TreeCharSet class is a helper class to streamline different line styles in
the printing routine of `Node`. `big_sep`, `med_sep` and `sma_sep` are used in
legacy printing routines of `Element`, `Molecule` and `Mixture`.
"""

from __future__ import annotations

from typing import TypeAlias, Any, Union, TypeVar, Literal
from collections.abc import Iterable
from dataclasses import dataclass


@dataclass(frozen=True)
class TreeCharSet:
    """Class for tree structure line ploting of class Node.
    
    It contains a set of characters for vertical lines, intersection lines,
    horizontal lines and angled lines (used for the last element of each level).
    """

    c_ver: str  # character vertical
    c_int: str  # character intersection
    c_hor: str  # character horizontal
    c_ang: str  # character angled

    @property
    def inter(self) -> str:
        """'Intersection' string for nodes that have further siblings."""
        return self.c_int + 3*self.c_hor
    
    @property
    def last(self) -> str:
        """'Last' string for nodes that have no further siblings (are the last child)."""
        return self.c_ang + 3*self.c_hor
    
    @property
    def fill(self) -> str:
        """'Filler' string to connect a node at a deeper level."""
        return 4*self.c_hor

    @property
    def vert(self) -> str:
        """'Vertical' string as horizontal filler for a child being printed at a deeper level."""
        return self.c_ver + 3*" "
    
    @property
    def empty(self) -> str:
        """'Empty' string as a horizontal filler for a child being printed at a deeper level for the last element."""
        return 4*" "

treeCharSets = {
    "basic": TreeCharSet("|", "+", "-", "+"),
    "box_drawings_light": TreeCharSet("\u2502", "\u251c", "\u2500", "\u2514"),
}
char_sets: TypeAlias = Literal["basic"] | Literal["box_drawings_light"]


Content: TypeAlias = Union[TypeVar("Substance"), TypeVar("Isotope")]

class Node:
    """Class to represent `Substance` and `Isotope` in hierarchical structure.
    
    Each node has a unique id, that is either the n-th created object (counter
    for class) or can be set manually. Printed is the label, that has to be set
    manually. Each node can have a parent and multiple children, representing
    the hierarchy and placing the node at a certain depth of it. Right alligned
    nodes will be printed at the maximum depth. Furthermore, each node can have
    content assigned to it, which is either of type `Substance` or `Isotope`.
    Optionally data may be specified in a dictionary.
    """

    _nodes = 0  # counter

    def __init__(self, label: str, parent: Node = None, content: Content = None,
                 id: Any = None, data: dict[str, float] = None
        ) -> None:
        """Constructor of node.

        Args:
            label:
                Descriptive name.
            parent:
                Parent node.
            content:
                Associated substance or isotope.
            id:
                Unique identifier.
            data:
                Associated data, containing additional information of the
                content.
        """
        
        self._nodes += 1

        self._id: Any
        self._label = label
        self._content: Content = content
        
        self._parent: Node | None = None
        self._depth: int
        self._children: list[Node] = []
        if data:
            self._data = data
        else:
            self._data = {}

        self._right_align = None

        if id is None:
            self._id = self._nodes
        else:
            self._id = id

        self.set_parent(parent)


    # ########
    # Properties
    # ########

    @property
    def id(self):
        """Unique Identifier."""
        return self._id
    
    @property
    def label(self):
        """Descriptive label."""
        return self._label
    
    @property
    def content(self):
        """Associated content."""
        return self._content
    
    @property
    def parent(self):
        """Parent node."""
        return self._parent
    
    @property
    def depth(self):
        """Depth of node in hierarchy."""
        return self._depth

    @property
    def children(self):
        """Children nodes."""
        return self._children
    
    @property
    def data(self):
        """Dictionary with additional data about content."""
        return self._data
    

    # ########
    # Construction
    # ########

    def set_parent(self, new_parent: Node | None) -> None:
        """Sets parent of node."""

        if self._parent is not None:
            self._parent._children.remove(self)  # remove itself from old parent

        self._parent = new_parent
        if new_parent is not None:
            new_parent._children.append(self)
        
        self._depth = self.count_depth()  # calculate depth

    def add_children(self, *args) -> None:
        """Adds children to node."""

        for arg in args:
            if isinstance(arg, Iterable):
                self.add_children(arg)
            elif isinstance(arg, Node):
                arg.set_parent(self)
            else:
                self.__class__(str(arg), content=arg, parent=self)

    def add_data(self, key: str, value: float) -> None:
        """Adds key-value (str-float) pair to data of node."""
        self._data[key] = value

    def align_right(self, level: int = -1, label_size: int = -1) -> None:
        """Set right alignment of node.

        Numbering of level is similar to list indexing. Positive values are
        absolute positions, negative values are relative to the end (max depth),
        e.g. "-1" is the maximum depth, "-2" the one before, etc. The aligment
        will still be at least the depth of the node.
        
        Args:
            level:
                Level of alignment.
            label_size:
                Fix size of node label, for better alignment.
        """

        self._right_align = level
        if label_size > 0:
            self._label = f"{self._label:>{label_size}}"

    
    # ########
    # Analysis
    # ########

    def is_root(self) -> bool:
        """Returns if node is root node (first)."""

        if self._parent is None:
            return True
        else:
            return False
    
    def is_leaf(self) -> bool:
        """Returns if node is leaf node (last)."""

        if not self._children:
            return True
        else:
            return False
        
    def count_depth(self) -> int:
        """Returns depth of node by counting parents, starting from 0."""
        
        if self._parent is None:
            return 0
        else:
            depth = 0
            cur_parent = self._parent
            while cur_parent is not None:
                 depth += 1
                 cur_parent = cur_parent._parent
            return depth
    
    def max_depth(self) -> int:
        """Returns maximal depth of node structure."""
        return max(node.depth for node in self)


    # ########
    # Fetch
    # ########

    def get_nodes_by_label(self, target: str) -> list[Node]:
        """Returns all nodes, which label equals target.
        
        Args:
            target:
                Label to search for.
                
        Returns:
            List of nodes which label matches target label.
        """
        return [node for node in self if node._label == target]

    def get_nodes_by_content(self, target: Content) -> list[Node]:
        """Returns all nodes, which content equals target.
        
        Args:
            target:
                Content object to search for.
                
        Returns:
            List of nodes which content matches target content.
        """
        return [node for node in self if node._content == target]


    # ########
    # Print
    # ########

    def _print_depth(self, max_depth: int) -> int:
        """Return level, whre node shall be printed.
        
        Args:
            max_depth:
                Maximum depth in tree.

        Returns:
            Level, where node shall be printed.
        """

        if self._right_align is None:
            return self._depth
        else:
            if self._right_align > 0:
                return max(self._depth, self._right_align)
            elif self._right_align < 0:
                return max(self._depth, max_depth+(self._right_align+1))


    def print_tree(
            self, char_set: char_sets = "box_drawings_light", *, 
            frac_fmt: str = "8.4f", prop_fmt: str = ".4f"
        ) -> None:
        """Prints node structure as tree.
        
        Args:
            char_set:
                Character set that is used to print lines in the tree.
            frac_fmt:
                Format string to be used for fractions.
            prop_fmt:
                Format string to be used for physical properties.
        """

        def print_node(node: Node):

            # construct prefix string
            pre = []
            if node._depth > 0:
                depth_parent = node._parent._print_depth(max_depth)

                # iterate over last_ones arry for structure
                for last in last_ones[:node._depth-1]:
                    if last == True:
                        pre.append(tcs.empty)
                    else:
                        pre.append(tcs.vert)
                if last_ones[node._depth-1] == True:
                    pre.append(tcs.last)
                else:
                    pre.append(tcs.inter)

                # empty buffers in front, when parent was already moved
                while len(pre) <= depth_parent:
                    pre.insert(-1, tcs.empty)

                # fill buffers at end, to reach right alignment
                while len(pre) < node._print_depth(max_depth):
                    pre.append(tcs.fill)

            pre = "".join(pre)[:-1]
            if pre:
                pre = pre + " "

            # construct data string
            data_str = []
            for key, value in node.data.items():
                if key == "x":
                    data_str.append(f"{value*1e2:{frac_fmt}} at.%")
                if key == "w":
                    data_str.append(f"{value*1e2:{frac_fmt}} wt.%")
                if key == "phi":
                    data_str.append(f"{value*1e2:{frac_fmt}} vol.%")
                if key == "M":
                    data_str.append(f"{value:{prop_fmt}} g/mol")
                if key == "rho":
                    data_str.append(f"{value:{prop_fmt}} g/cm^3")
            data_str =  "  |  ".join(data_str)

            # actual print of current node
            label_str = f"\"{node.label.strip()}\""
            label_str_len = len(node.label) + 2
            print(f"{pre}{node.content.__class__.__name__} {label_str:>{label_str_len}}: {data_str}")

            # print children
            if node._children:
                last_ones[node._depth] = False
                for child in node._children[:-1]:
                    print_node(child)
                else:
                    last_ones[node._depth] = True
                print_node(node._children[-1])

        # get tree character set
        try:
            tcs = treeCharSets[char_set]
        except KeyError as ex:
            print("Unknown character set for tree plotting. Instead using default value.")
            tcs = treeCharSets["box_drawings_light"]

        # setup print variables
        max_depth = self.max_depth()
        last_ones = [False for i in range(max_depth)]

        # start recursive printing routine
        print_node(self)


    # ########
    # Operators
    # ########

    def __str__(self) -> str:
        return f"{self._label}"
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__} of \"{repr(self._content)}\""
    
    def __eq__(self, other) -> bool:
        if type(self) is type(other):
            return self._id == other._id
        else:
            return NotImplemented
    
    def __hash__(self) -> int:
        return hash((self.__class__, self._id))
    
    def __iter__(self):
        if self.is_root():
            yield self
        for child in self._children:
            yield child
            yield from child.__iter__()
        


"""Class for Element

Element class serves as constituents for `Molecule`.
"""

from .substance import Substance
from .isotope import Isotope
from .isotopes import _natural_compositions
from .constants import ATOM_NUMB_TO_SYMBOL


class Element(Substance):
    """A substance containing only isotopes with the same atomic number.

    An element shares its properties with its parent class `Substance`, with
    the exception that constituents can only be isotopes and therefore has a
    well-defined atomic number Z and relative atomic mass A_r itself. An 
    element can be flagged as a natural element, indicating the natural 
    abundance of its isotopes. 
    """

    # override
    @classmethod
    def _get_allowed_constituents(cls):
        return (Isotope,)
    
    _RIGHT_ALIGN_PREF = -2  # override

    def __init__(self, name: str, composition: dict[Isotope, float], mode: str = "_legacy", natural: bool = False, **kwargs) -> None:
        """Constructor of element.
        
        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps an isotope of the element to its
                fraction. The fractions are physically interpreted according
                to the given mode. Values are normalsied and don't need to add
                up to unity.
            mode:
                Fractions in the composition can be interpreted as atomic or
                weight fractions. The default value allows inputs as before
                v1.1.0, where positive fractions refer to atomic and negative
                fractions to weight.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.

        Raises:
            ValueError:
                If non-valid constructor mode or given constituent is not an
                isotope.
        """

        super().__init__(name=name, composition=composition, mode=mode, **kwargs)

        self._is_natural = natural

        # take atomic number of first isotope (and make sure all isotopes are of the same element)
        self._Z: int = next(iter(composition)).Z  # atomic number of element
        if not all(self._Z == isotope.Z for isotope in composition.keys()):
            raise ValueError(f"Atomic number of all isotopes of {self.__class__.__name__} \"{self._name}\" must match!")
        
        # construct symbol
        if not self._symbol:
            self._symbol = self.element_symbol()
        

    # override
    @classmethod
    def _inp_composition_volume(cls, inp_composition: dict[Isotope, float]) -> dict[Isotope, float]:
        raise ValueError(f"Input mode 'volume' is disabled for {cls.__name__} creation.")


    # ########
    # Properties
    # ########

    @property
    def Z(self):
        """Atomic number of the element."""
        return self._Z
    
    @property
    def A_r(self):
        """Relative atomic mass (atomic weight) [-] of the element."""
        return self._calc_A_r()


    # ########
    # Quantity Calculation
    # ########

    def _calc_A_r(self) -> float:
        r"""Calculates relative atomic mass (atomic weight) [-] of the element.
        
        The relative atomic masses of all isotopes are weighted by their atomic
        fraction and summed up:
            $$\overline{M} = \sum_i \left( x_i \cdot A_\mathrm{r},i \right)$$
        """
        return sum(x_i*isotope.A_r for isotope, x_i in self._composition.items())


    # ########
    # Collection
    # ########

    def surrogate_isotope(self) -> Isotope | None:
        """Returns the surrogate isotope for a natural element or None if not natural."""
        if self._is_natural:
            return _natural_compositions[self._Z]
        else:
            return None

    # override
    def _append_elements(self, element_list: dict, by_weight: bool = False, f_p: float = 1.0):

        element_list[-1] = element_list[-1] + 1  # increase counter
        element_list[element_list[-1]] = (self, f_p)  # append element
        element_list[-1] = element_list[-1] + len(self._composition)  # skip numbers of isotopes in element

        return element_list


    # ########
    # Functions
    # ########

    def element_symbol(self) -> str:
        """Returns the element symbol as a string."""
        return ATOM_NUMB_TO_SYMBOL[self._Z]


    # ########
    # Operators
    # ########
    
    def __hash__(self):
        #return hash((self.__class__, self._name))
        return hash((self.__class__, self._Z, self._M))

    def __eq__(self, other):
        try:
            if (self._Z == other._Z) and (self._M == other._M):
                return True
            else:
                return False
        except:
            return NotImplemented
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        try:
            if self._Z < other._Z:
                return True
            elif self._Z == other._Z:
                if self._M < other._M:
                    return True
                else:
                    return False
            else:
                return False
        except:
            return NotImplemented

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other):
        try:
            if self._Z > other._Z:
                return True
            elif self._Z == other._Z:
                if self._M > other._M:
                    return True
                else:
                    return False
            else:
                return False
        except:
            return NotImplemented

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)


"""Class for Molecule

Molecule class serves as constituents for `Mixture`.
"""

from .substance import Substance
from .element import Element


class Molecule(Substance):
    """A substance containing at least two atoms, potentially from different elements.

    A molecule shares its properties with its parent class `Substance`, with
    the exception that constituents can only be elements. The distribution of
    elements can be expressed as fractions or in integers, which sum up to the
    total amount of atoms in the molecule.
    """

    # override
    @classmethod
    def _get_allowed_constituents(cls):
        return (Element,)

    def __init__(self, name: str, composition: dict[Element, float], mode: str = "_legacy", **kwargs) -> None:
        """Constructor of molecule.
        
        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps an element of the molecule to its atomic
                fraction or number of atoms. Values are normalsied and don't
                need to add up to unity.
            mode:
                Fractions in elements can be interpreted only as atomic
                fractions or number of atoms. The default value allows inputs
                as before 1.1.0.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.

        Raises:
            ValueError:
                If non-valid constructor mode or given constituent is not an
                element.
        """

        self._atoms: int = int(sum(composition.values()))  # number of atoms in molecule

        super().__init__(name=name, composition=composition, mode=mode, **kwargs)

        # construct symbol
        if not self._symbol:
            sym = ""
            for element, atoms in self.get_composition_in_atoms().items():
                sym += f"{element.element_symbol()}"
                if atoms > 1:
                    sym += f"{atoms:.2g}"
            self._symbol = sym
        
    # override
    @classmethod
    def _inp_composition_weight(cls, inp_composition: dict[Element, float]) -> dict[Element, float]:
        raise ValueError(f"Input mode 'weight' is disabled for {cls.__name__} creation.")
    
    # override
    @classmethod
    def _inp_composition_volume(cls, inp_composition: dict[Element, float]) -> dict[Element, float]:
        raise ValueError(f"Input mode 'volume' is disabled for {cls.__name__} creation.")


    # ########
    # Properties
    # ########

    @property
    def atoms(self):
        """Number of atoms in the molecule."""
        return self._atoms
    

    # ########
    # Quantity Calculation
    # ########

    # override
    def _calc_M(self) -> float:
        r"""Calculates molar mass.
        
        The molar masses of all constituents are multiplied by their number of
        atoms in the molecule and summed up:
            $$\overline{M} = \sum_i \left( N_i \cdot M_i \right)$$
        Will return zero if calculation is not possible.
        """
        if all(constituent.M > 0 for constituent in self._composition.keys()):
            return sum(N_i*constituent.M for constituent, N_i in self.get_composition_in_atoms().items())
        else:
            return 0.0


    # ########
    # Conversion
    # ########

    def get_composition_in_atoms(self) -> dict[Element, float]:
        """Returns constituents with their number of atoms."""
        return {element: int(x_i*self._atoms) for element, x_i in self._composition.items()}


    # ########
    # Collection
    # ########

    # override
    def _append_elements(self, element_list: dict[int, tuple[Substance, float]], by_weight: bool = False, f_p: float = 1.0):

        element_list[-1] = element_list[-1] + 1  # increase counter

        if not by_weight:
            composition = self.get_composition_in_atoms()
        else:
            composition = self.get_composition_in_wt()  
        
        for constituent, f_i in composition.items():
            constituent._append_elements(element_list, by_weight, f_p*f_i)

        return element_list



"""Functions for parts-per-notation and conversion of atomic, weight and volume percent.

For interpreting the LaTeX equations, https://quicklatex.com/ could be used for quick access.
"""


# to weight fractions (w_i)

def at_to_wt(at_fracs: list[float], molar_masses: list[float]) -> list[float]:
    r"""Converts atomic fractions to weight fractions.
    
    $$w_i = \frac{x_i \cdot M_i}{\sum_{z=1}^{Z} \left( x_z \cdot M_z \right)}$$

    Args:
        at_fracs:
            Atomic fractions.
        molar_masses:
            Corresponding values of molar masses.
    
    Returns:
        Corresponding values of weight fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for molar masses.
    """

    if len(at_fracs) == len(molar_masses):  # same dimensions
        if all(molar_masses):  # no zero molar masses
            
            # calculate (constant) denominator for conversion
            denominator = sum(x_z * M_z for x_z, M_z in zip(at_fracs, molar_masses))

            # convert to weight fraction
            wt_fracs = [(x_i*M_i) / denominator for x_i, M_i in zip(at_fracs, molar_masses)]

            return wt_fracs

        else:
            raise ValueError(f"One of the constituents does not have a molar mass.")

    else:
        raise ValueError(f"Lists of weight fractions and molar masses must have the same length.")

def vol_to_wt(vol_fracs: list[float], densities: list[float]) -> list[float]:
    r"""Converts volume fractions to weight fractions.
    
    $$w_i = \frac{\phi_i \cdot \rho_i}{\sum_{z=1}^{Z} \left( \phi_z \cdot \rho_z \right)}$$

    Args:
        vol_fracs:
            Volume fractions.
        densities:
            Corresponding values of densities.
    
    Returns:
        Corresponding values of weight fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for densities.
    """

    if len(vol_fracs) == len(densities):  # same dimensions
        if all(densities):  # no zero densities
            
            # calculate (constant) denominator for conversion
            denominator = sum(phi_z * rho_z for phi_z, rho_z in zip(vol_fracs, densities))

            # convert to weight fraction
            wt_fracs = [(phi_i*rho_i) / denominator for phi_i, rho_i in zip(vol_fracs, densities)]

            return wt_fracs

        else:
            raise ValueError(f"One of the constituents does not have a density.")

    else:
        raise ValueError(f"Lists of weight fractions and molar masses must have the same length.")


# to atomic (mole) fractions (x_i)

def wt_to_at(wt_fracs: list[float], molar_masses: list[float]) -> list[float]:
    r"""Converts weight fractions to atomic (mole) fractions.
    
    $$x_i = \frac{w_i / M_i}{\sum_{z=1}^{Z} \left( w_z / M_z \right)}$$

    Args:
        wt_fracs:
            Weight fractions.
        molar_masses:
            Corresponding values of molar masses.
    
    Returns:
        Corresponding values of atomic fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for molar masses.
    """

    if len(wt_fracs) == len(molar_masses):  # same dimensions
        if all(molar_masses):  # no zero molar masses
            
            # calculate (constant) denominator for conversion
            denominator = sum(w_z / M_z for w_z, M_z in zip(wt_fracs, molar_masses))

            # convert to atomic fraction
            at_fracs = [(w_i/M_i) / denominator for w_i, M_i in zip(wt_fracs, molar_masses)]

            return at_fracs

        else:
            raise ValueError(f"One of the constituents does not have a molar mass.")

    else:
        raise ValueError(f"Lists of weight fractions and molar masses must have the same length.")

def vol_to_at(vol_fracs: list[float], molar_volumes: list[float]) -> float:
    r"""Converts volume fractions to atomic (mole) fractions.

    $$x_i = \frac{\varphi_i / V_{\mathrm{m},i}}{\sum_{z=1}^{Z} \left( \varphi_z / V_{\mathrm{m},z} \right)}
          = \frac{\varphi_i \cdot \rho_i / M_i}{\sum_{z=1}^{Z} \left( \varphi_z \cdot \rho_z / M_z \right)}$$

    Args:
        vol_fracs:
            Volume fractions.
        molar_volumes:
            Corresponding values of molar volumes.
    
    Returns:
        Corresponding values of atomic fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for molar volumes.
    """

    if len(vol_fracs) == len(molar_volumes):  # same dimensions
        if all(molar_volumes):  # no zero molar volumes
            
            # calculate (constant) denominator for conversion
            denominator = sum(phi_z / V_m_z for phi_z, V_m_z in zip(vol_fracs, molar_volumes))

            # convert to atomic fraction
            at_fracs = [(phi_i/V_m_i) / denominator for phi_i, V_m_i in zip(vol_fracs, molar_volumes)]

            return at_fracs

        else:
            raise ValueError(f"One of the constituents does not have a molar volume.")

    else:
        raise ValueError(f"Lists of volume fractions and molar volumes must have the same length.")


# to volume fractions (phi_i)

def wt_to_vol(wt_fracs: list[float], densities: list[float]) -> list[float]:
    r"""Converts weight fractions to volume fractions.
    
    $$\varphi_i = \frac{w_i / \rho_i}{\sum_{z=1}^{Z} \left( w_z / \rho_z \right)}$$

    Args:
        wt_fracs:
            Weight fractions.
        densities:
            Corresponding values of densities.
    
    Returns:
        Corresponding values of volume fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for densities.
    """

    if len(wt_fracs) == len(densities):  # same dimensions
        if all(densities):  # no zero densities
            
            # calculate (constant) denominator for conversion
            denominator = sum(w_z / rho_z for w_z, rho_z in zip(wt_fracs, densities))

            # convert to volume fraction
            vol_fracs = [(w_i/rho_i) / denominator for w_i, rho_i in zip(wt_fracs, densities)]

            return vol_fracs

        else:
            raise ValueError(f"One of the constituents does not have a density.")

    else:
        raise ValueError(f"Lists of weight fractions and densities must have the same length.")

def at_to_vol(at_fracs: list[float], molar_volumes: list[float]) -> list[float]:
    r"""Converts atomic fractions to volume fractions.
    
    $$\varphi_i = \frac{x_i \cdot V_{\mathrm{m},i}}{\sum_{z=1}^{Z} \left( x_z \cdot V_{\mathrm{m},z} \right)}
                = \frac{x_i \cdot M_i / \rho_i}{\sum_{z=1}^{Z} \left( x_z \cdot M_z / \rho_z \right)}$$
    
    Args:
        at_fracs:
            Atomic fractions.
        molar_volumes:
            Corresponding values of molar volumes.
    
    Returns:
        Corresponding values of volume fractions.

    Raises:
        ValueError: If lists have different lengths or zero values for molar volumes.
    """

    if len(at_fracs) == len(molar_volumes):  # same dimensions
        if all(molar_volumes):  # no zero molar masses
            
            # calculate (constant) denominator for conversion
            denominator = sum(x_z * V_m_z for x_z, V_m_z in zip(at_fracs, molar_volumes))

            # convert to volume fraction
            wt_fracs = [(x_i*V_m_i) / denominator for x_i, V_m_i in zip(at_fracs, molar_volumes)]

            return wt_fracs

        else:
            raise ValueError(f"One of the constituents does not have a molar volume.")

    else:
        raise ValueError(f"Lists of atomic (mole) fractions and molar volumes must have the same length.")




# parts-per functions

def percent(value: float):
    """Converts value from percent."""
    return value*1e-2
perc = percent  # function alias
pc   = percent  # function alias

def permille(value: float):
    """Converts value from per mille."""
    return value*1e-3
pm = permille  # function alias

def permyriad(value: float):
    """Converts value from per myriad."""
    return value*1e-4
bp = permyriad  # function alias

def percentmille(value: float):
    """Converts value from per cent mille."""
    return value*1e-5
pcm = percentmille  # function alias

def ppm(value: float):
    """Converts value from parts per million."""
    return value*1e-6

def ppb(value: float):
    """Converts value from parts per billion."""
    return value*1e-9

def ppt(value: float):
    """Converts value from parts per trillion."""
    return value*1e-12

def ppq(value: float):
    """Converts value from parts per quadrillion."""
    return value*1e-15


"""Class for Isotope.

Isotope class serves as constituent for Element class.
"""

from __future__ import annotations

from .constants import ATOM_NUMB_TO_SYMBOL, M_u


class Isotope:
    """Isotope are nuclides of the same element.

    Each isotope is characterised by their atomic number (Z), mass number (M)
    and isomeric state (I). Furthermore, each isotope has a chracteristic 
    relative atomic mass (A_R). They are used to compose elements. A name can
    be given, otherwise a name is build from the symbol of the corresponding
    element and the mass number of the isotope.
    """

    def __init__(self, Z: int, A: int, A_r: float, I: int = 0, name: str = "") -> None:
        """Constructor of isotope.
        
        Args:
            Z:
                Atomic number (number of protons).
            A:
                Mass number (number of protons + neutrons).
            A_r:
                Relative atomic mass ("atomic weight") [-].
            I:
                Isomeric state.
            name:
                Override name of isotope.
        """
        
        self._Z: int = Z        # atomic number (protons)
        self._A: int = A        # mass number (protons + neutrons)
        self._A_r: float = A_r  # relative atomic mass (atomic weight) [-]
        self._I: int = I        # isomeric state (0 = ground state, 1 = metastable, ...)
        self._name: str         # name

        if name == "":
            self._name = self.short_name()
        else:
            self._name = name


    # ########
    # Getter
    # ########

    @property
    def name(self):
        """Descriptive name."""
        return self._name

    @property
    def Z(self):
        """Atomic number (number of protons)."""
        return self._Z

    @property
    def A(self):
        """Mass number (number of protons + neutrons)."""
        return self._A

    @property
    def A_r(self):
        """Relative atomic mass (atomic weight) [-]."""
        return self._A_r

    @property
    def I(self):
        """Isomeric state."""
        return self._I

    @property
    def M(self):
        """Molar mass [g mol^-1]."""
        return self.calc_M()

    @property
    def N(self):
        """Neutron number (number of neutrons)."""
        return self.calc_N()
    
    @property
    def ZA(self):
        """ZA notation."""
        return self.ZA_notation()
    
    @property
    def ZAI(self):
        """ZAI notation."""
        return self.ZAI_notation()


    # ########
    # Functions
    # ########

    def calc_M(self) -> float:
        """Calculates molar mass [g mol^-1] of the isotope."""
        return self._A_r * M_u * 1e+03
    
    def calc_N(self) -> int:
        """Calculates neutron number (number of neutrons)."""
        return self._A - self._Z
    
    def element_symbol(self) -> str:
        """Returns symbol of element that isotope represents."""
        return ATOM_NUMB_TO_SYMBOL[self._Z]
    
    def ZA_notation(self) -> int:
        """Returns the ZA notation of the isotope.
        
        ZA = Z*1000 + A
        """
        return self._Z*1000 + self._A
    
    def ZAI_notation(self) -> int:
        """Returns the ZAI notation of the isotope.

        ZAI = Z*10000 + A*10 + I
        """
        return self._Z*10000 + self._A*10 + self._I

    def short_name(self, verbose: bool = False) -> str:
        """Returns short name of isotope as string.
        
        Similar to pronunciation of AZE notation.

        Args:
            verbose:
                Flag to explicitly show isomeric state of isotope.
        """
        
        name = f"{self.element_symbol()}-{self._A}"
        
        if verbose:
            if self._I == 0:
                name += "g"
            else:
                name += f"m{self._I}"
        else:
            if self._I == 1:
                name += "m"
            if self._I > 1:
                name += f"m{self._I}"
            
        return name
    

    # ########
    # Print
    # ########



    # ########
    # Operators
    # ########

    def __str__(self):
        return self._name

    def __repr__(self):
        return f"{self.__class__.__name__} \"{self._name}\""

    def __hash__(self):
        return hash((self.__class__, self._Z, self._A, self._I))

    def __eq__(self, other):
        try:
            if (self._Z == other._Z) and (self._A == other._A) and (self._I == other._I):
                return True
            else:
                return False
        except:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        try:
            if self._Z < other._Z:
                return True
            elif self._Z == other._Z:
                if self._A < other._A:
                    return True
                elif self._A == other._A:
                    if self._I < other._I:
                        return True
                    else:
                        return False
                else:
                    return False
            else:
                return False
        except:
            return NotImplemented

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)


    def __gt__(self, other):
        try:
            if self._Z > other._Z:
                return True
            elif self._Z == other._Z:
                if self._A > other._A:
                    return True
                elif self._A == other._A:
                    if self._I > other._I:
                        return True
                    else:
                        return False
                else:
                    return False
            else:
                return False
        except:
            return NotImplemented

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)




"""Class for Mixture.

Mixture class is the top class, but may serve as a constituent for itself.
"""

from typing import Union, TypeAlias

from .substance import Substance
from .element import Element
from .molecule import Molecule


Constituent: TypeAlias = Union["Mixture", Molecule, Element]

class Mixture(Substance):
    """A substance made of elements, molecules or other mixtures.

    A molecule shares its properties with its parent class `Substance`.
    Furthermore, it is possible to calculate the density of a mixture, as long
    as all its constituents have a specified density.
    """

    # override
    @classmethod
    def _get_allowed_constituents(cls):
        return (Mixture, Molecule, Element)
    
    def __init__(self, name: str, composition: dict[Constituent, float], mode: str = "_legacy", **kwargs) -> None:
        """Constructor of mixture.
        
        Args:
            name:
                Descriptive name.
            composition:
                Dictionary that maps a constituent of the mixture to its
                fraction. The fractions are physically interpreted according
                to the given mode. Values are normalsied and don't need to add
                up to unity.
            mode:
                Fractions in the composition can be interpreted as atomic,
                weight or volumetric fractions. Volumetric is only valid, if
                all constituents have a density. The default value allows
                inputs as before v1.1.0, where positive fractions refer to
                atomic and negative fractions to weight.
            **kwargs:
                Keyword arguments to override values.
        
        Keyword Args:
            M (float):
                Set molar mass.
            rho (float):
                Set density.
            symbol (str):
                Short symbol.
            data (dict[str, float]):
                Dictionary with additional data to store or use for mean values
                in mixtures.

        Raises:
            ValueError:
                If non-valid constructor mode or given constituent is not a
                substance.
        """

        super().__init__(name=name, composition=composition, mode=mode, **kwargs)

        if self._rho == 0.0:
            self._rho = self._calc_rho()


    # ########
    # Quantity Calculation
    # ########

    def _calc_rho(self) -> float:
        r"""Calculates average density.
        
        The mean of all constituent densities, weighted by their volume fraction:
            $$\overline{\rho} = \sum_i \phi_i \cdot \rho_i$$
        Returns zero if calculation is not possible.
        """
        try:
            return self.rule_of_mixtures("volume", "density")
        except:
            return 0.0
        
    def rule_of_mixtures(self, mode: str, quantity: str, inverse: bool = False) -> float:
        r"""Calculates average quantity value for a mixture.

        Average value is either calculated via general rule of mixtures:
            $$\overline{\psi} = \sum_i f_i \cdot \psi_i$$
        or if selected via the inverse rule of mixtures:
            $$\overline{\psi} = \left( \sum_i \frac{f_i}{\psi_i} \right)^{-1}$$
        
        Args:
            mode:
                Composition fractions used for weighting. Possible fraction
                modes are `atomic`/`mole`, `weight` or `volume` Volumetric 
                fractions are only valid, if all constituents have a density.
            quantity:
                Quantity, which should be calculated as average.
            inverse:
                Use inverse rule of mixtures.

        Returns:
            Calculated mean value.

        Raises:
            ValueError:
                If a constituent contains a zero value for relevant quantities
                or fraction mode is unknown.
            KeyError:
                If a constituent does not contain the key (and thus data) for
                the selected quantity.
        """

        # get composition in selected mode
        if mode in {"atomic", "at", "mole", "mol"}:
            composition = self._composition
        elif mode in {"weight", "wt"}:
            if all(constituent.M > 0 for constituent in self._composition.keys()):
                composition = self.get_composition_in_wt()
            else:
                raise ValueError("One of the constituents contains zero-values for molar mass.")
        elif mode in {"volume", "vol"}:
            if all((constituent.M > 0 and constituent.rho > 0) for constituent in self._composition.keys()):
                composition = self.get_composition_in_vol()
            else:
                raise ValueError("One of the constituents contains zero-values for molar mass or density.")
        else:
            raise ValueError(f"Unknown fraction mode \"{mode}\".")
        
        # extract fractions
        fractions = composition.values()

        # extract quantities
        if quantity == "density" or quantity == "rho":
            quantities = [constituent.rho for constituent in composition.keys()]
        else:
            try:
                quantities = [constituent.data[quantity] for constituent in composition.keys()]
            except KeyError:
                raise KeyError("One of the constituents does not contain values for the selected quantity.")

        # check if all quantities are non-zero
        if any(quantity == 0 for quantity in quantities):
            raise ValueError("One of the constituents contains zero-values for the selected quantity.")
        
        if inverse:
            return self._inverse_rule_of_mixtures(fractions, quantities)
        else:
            return self._general_rule_of_mixtures(fractions, quantities)    

    @staticmethod
    def _general_rule_of_mixtures(fractions: list[float], quantities: list[float]) -> float:
        r"""Calculates average value for a quantity.

        Weighted average of quantities with given fractions:
            $$\overline{\psi} = \sum_i f_i \cdot \psi_i$$

        Args:
            fractions:
                List with fractions of each constituent.
            quantities:
                List with quantity values of each constituent.
        """
        return sum(fraction*quantity for fraction, quantity in zip(fractions, quantities))

    @staticmethod
    def _inverse_rule_of_mixtures(fractions: list[float], quantities: list[float]) -> float:
        r"""Calculates average value for a quantity.

        Inverse of weighted average of quantities with given fractions:
            $$\overline{\psi} = \left( \sum_i \frac{f_i}{\psi_i} \right)^{-1}$$
        
        Args:
            fractions:
                List with fractions of each constituent.
            quantities:
                List with quantity values of each constituent.
        """
        return sum(fraction/quantity for fraction, quantity in zip(fractions, quantities))**-1